import os
import platform

import pytest
from uqbar.strings import normalize

from supriya import SynthDef
from supriya.ugens import SuperColliderSynthDef, decompile_synthdef
from supriya.ugens.system import (
    amplitude_scope_audio_1,
    amplitude_scope_control_2,
    frequency_scope_lin_1,
    frequency_scope_lin_shm_1,
    frequency_scope_log_1,
    frequency_scope_log_shm_2,
    system_link_audio_1,
    system_link_audio_2,
)


@pytest.mark.parametrize(
    "synthdef, expected_str",
    [
        (
            system_link_audio_1,
            """
            synthdef:
                name: supriya:link-ar:1
                ugens:
                -   Control.kr:
                        done_action: 2.0
                        fade_time: 0.02
                        gate: 1.0
                        in_: 16.0
                        out: 0.0
                -   BinaryOpUGen(LESS_THAN_OR_EQUAL).kr:
                        left: Control.kr[1:fade_time]
                        right: 0.0
                -   EnvGen.kr:
                        gate: Control.kr[2:gate]
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: Control.kr[1:fade_time]
                        done_action: Control.kr[0:done_action]
                        envelope[0]: BinaryOpUGen(LESS_THAN_OR_EQUAL).kr[0]
                        envelope[1]: 2.0
                        envelope[2]: 1.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 1.0
                        envelope[6]: 3.0
                        envelope[7]: 0.0
                        envelope[8]: 0.0
                        envelope[9]: 1.0
                        envelope[10]: 3.0
                        envelope[11]: 0.0
                -   InFeedback.ar:
                        channel_count: 1
                        bus: Control.kr[3:in_]
                -   BinaryOpUGen(MULTIPLICATION).ar:
                        left: InFeedback.ar[0]
                        right: EnvGen.kr[0]
                -   Out.ar:
                        bus: Control.kr[4:out]
                        source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
            """,
        ),
        (
            system_link_audio_2,
            """
            synthdef:
                name: supriya:link-ar:2
                ugens:
                -   Control.kr:
                        done_action: 2.0
                        fade_time: 0.02
                        gate: 1.0
                        in_: 16.0
                        out: 0.0
                -   BinaryOpUGen(LESS_THAN_OR_EQUAL).kr:
                        left: Control.kr[1:fade_time]
                        right: 0.0
                -   EnvGen.kr:
                        gate: Control.kr[2:gate]
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: Control.kr[1:fade_time]
                        done_action: Control.kr[0:done_action]
                        envelope[0]: BinaryOpUGen(LESS_THAN_OR_EQUAL).kr[0]
                        envelope[1]: 2.0
                        envelope[2]: 1.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 1.0
                        envelope[6]: 3.0
                        envelope[7]: 0.0
                        envelope[8]: 0.0
                        envelope[9]: 1.0
                        envelope[10]: 3.0
                        envelope[11]: 0.0
                -   InFeedback.ar:
                        channel_count: 2
                        bus: Control.kr[3:in_]
                -   BinaryOpUGen(MULTIPLICATION).ar/0:
                        left: InFeedback.ar[0]
                        right: EnvGen.kr[0]
                -   BinaryOpUGen(MULTIPLICATION).ar/1:
                        left: InFeedback.ar[1]
                        right: EnvGen.kr[0]
                -   Out.ar:
                        bus: Control.kr[4:out]
                        source[0]: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                        source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
            """,
        ),
        (
            amplitude_scope_audio_1,
            """
            synthdef:
                name: supriya:amp-scope-ar:1
                ugens:
                -   Control.kr:
                        in_: 0.0
                        max_frames: 4096.0
                        scope_frames: 4096.0
                        scope_id: 0.0
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in_]
                -   ScopeOut2.ar:
                        scope_id: Control.kr[3:scope_id]
                        max_frames: Control.kr[1:max_frames]
                        scope_frames: Control.kr[2:scope_frames]
                        source[0]: In.ar[0]
            """,
        ),
        (
            amplitude_scope_control_2,
            """
            synthdef:
                name: supriya:amp-scope-kr:2
                ugens:
                -   Control.kr:
                        in_: 0.0
                        max_frames: 4096.0
                        scope_frames: 4096.0
                        scope_id: 0.0
                -   In.kr:
                        channel_count: 2
                        bus: Control.kr[0:in_]
                -   K2A.ar/0:
                        source: In.kr[0]
                -   K2A.ar/1:
                        source: In.kr[1]
                -   ScopeOut2.ar:
                        scope_id: Control.kr[3:scope_id]
                        max_frames: Control.kr[1:max_frames]
                        scope_frames: Control.kr[2:scope_frames]
                        source[0]: K2A.ar/0[0]
                        source[1]: K2A.ar/1[0]
            """,
        ),
        (
            frequency_scope_lin_1,
            """
            synthdef:
                name: supriya:freq-scope-lin:1
                ugens:
                -   Control.ir:
                        fft_buffer_size: 2048.0
                        rate: 4.0
                        scope_id: 0.0
                -   UnaryOpUGen(RECIPROCAL).ir:
                        source: Control.ir[0:fft_buffer_size]
                -   BinaryOpUGen(MULTIPLICATION).ir/0:
                        left: Control.ir[1:rate]
                        right: UnaryOpUGen(RECIPROCAL).ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir:
                        left: 1.0
                        right: BinaryOpUGen(MULTIPLICATION).ir/0[0]
                -   Control.kr:
                        in_: 0.0
                -   MaxLocalBufs.ir:
                        maximum: 1.0
                -   LocalBuf.ir:
                        channel_count: 1.0
                        frame_count: Control.ir[0:fft_buffer_size]
                -   BufSamples.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(MULTIPLICATION).ir/1:
                        left: BufSamples.ir[0]
                        right: 0.5
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in_]
                -   FFT.kr:
                        buffer_id: LocalBuf.ir[0]
                        source: In.ar[0]
                        hop: 0.75
                        window_type: 1.0
                        active: 1.0
                        window_size: 0.0
                -   PV_MagSmear.kr:
                        pv_chain: FFT.kr[0]
                        bins: 1.0
                -   BufDur.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir:
                        left: Control.ir[1:rate]
                        right: BufDur.ir[0]
                -   LFSaw.ar:
                        frequency: BinaryOpUGen(FLOAT_DIVISION).ir[0]
                        initial_phase: BinaryOpUGen(SUBTRACTION).ir[0]
                -   MulAdd.ar:
                        source: LFSaw.ar[0]
                        multiplier: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        addend: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                -   BinaryOpUGen(ROUND).ar:
                        left: MulAdd.ar[0]
                        right: 2.0
                -   BufRd.ar:
                        channel_count: 1
                        buffer_id: LocalBuf.ir[0]
                        phase: BinaryOpUGen(ROUND).ar[0]
                        loop: 1.0
                        interpolation: 1.0
                -   BinaryOpUGen(FLOAT_DIVISION).ar:
                        left: BufRd.ar[0]
                        right: Control.ir[0:fft_buffer_size]
                -   UnaryOpUGen(AMPLITUDE_TO_DB).ar:
                        source: BinaryOpUGen(FLOAT_DIVISION).ar[0]
                -   ScopeOut.ar:
                        buffer_id: Control.ir[2:scope_id]
                        source[0]: UnaryOpUGen(AMPLITUDE_TO_DB).ar[0]
            """,
        ),
        (
            frequency_scope_log_1,
            """
            synthdef:
                name: supriya:freq-scope-log:1
                ugens:
                -   Control.ir:
                        fft_buffer_size: 2048.0
                        rate: 4.0
                        scope_id: 0.0
                -   UnaryOpUGen(RECIPROCAL).ir:
                        source: Control.ir[0:fft_buffer_size]
                -   BinaryOpUGen(MULTIPLICATION).ir/0:
                        left: Control.ir[1:rate]
                        right: UnaryOpUGen(RECIPROCAL).ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir:
                        left: 1.0
                        right: BinaryOpUGen(MULTIPLICATION).ir/0[0]
                -   Control.kr:
                        in_: 0.0
                -   MaxLocalBufs.ir:
                        maximum: 1.0
                -   LocalBuf.ir:
                        channel_count: 1.0
                        frame_count: Control.ir[0:fft_buffer_size]
                -   BufSamples.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(MULTIPLICATION).ir/1:
                        left: BufSamples.ir[0]
                        right: 0.5
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in_]
                -   FFT.kr:
                        buffer_id: LocalBuf.ir[0]
                        source: In.ar[0]
                        hop: 0.75
                        window_type: 1.0
                        active: 1.0
                        window_size: 0.0
                -   PV_MagSmear.kr:
                        pv_chain: FFT.kr[0]
                        bins: 1.0
                -   BufDur.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir:
                        left: Control.ir[1:rate]
                        right: BufDur.ir[0]
                -   LFSaw.ar:
                        frequency: BinaryOpUGen(FLOAT_DIVISION).ir[0]
                        initial_phase: BinaryOpUGen(SUBTRACTION).ir[0]
                -   MulAdd.ar:
                        source: LFSaw.ar[0]
                        multiplier: 0.5
                        addend: 0.5
                -   BinaryOpUGen(POWER).ar:
                        left: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        right: MulAdd.ar[0]
                -   BinaryOpUGen(MULTIPLICATION).ar:
                        left: BinaryOpUGen(POWER).ar[0]
                        right: 2.0
                -   BinaryOpUGen(ROUND).ar:
                        left: BinaryOpUGen(MULTIPLICATION).ar[0]
                        right: 2.0
                -   BufRd.ar:
                        channel_count: 1
                        buffer_id: LocalBuf.ir[0]
                        phase: BinaryOpUGen(ROUND).ar[0]
                        loop: 1.0
                        interpolation: 1.0
                -   BinaryOpUGen(FLOAT_DIVISION).ar:
                        left: BufRd.ar[0]
                        right: Control.ir[0:fft_buffer_size]
                -   UnaryOpUGen(AMPLITUDE_TO_DB).ar:
                        source: BinaryOpUGen(FLOAT_DIVISION).ar[0]
                -   ScopeOut.ar:
                        buffer_id: Control.ir[2:scope_id]
                        source[0]: UnaryOpUGen(AMPLITUDE_TO_DB).ar[0]
            """,
        ),
        (
            frequency_scope_lin_shm_1,
            """
            synthdef:
                name: supriya:freq-scope-lin-shm:1
                ugens:
                -   Control.ir:
                        fft_buffer_size: 2048.0
                        rate: 4.0
                        scope_id: 0.0
                -   UnaryOpUGen(RECIPROCAL).ir:
                        source: Control.ir[0:fft_buffer_size]
                -   BinaryOpUGen(MULTIPLICATION).ir/0:
                        left: Control.ir[1:rate]
                        right: UnaryOpUGen(RECIPROCAL).ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir:
                        left: 1.0
                        right: BinaryOpUGen(MULTIPLICATION).ir/0[0]
                -   Control.kr:
                        in_: 0.0
                -   MaxLocalBufs.ir:
                        maximum: 1.0
                -   LocalBuf.ir:
                        channel_count: 1.0
                        frame_count: Control.ir[0:fft_buffer_size]
                -   BufSamples.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(MULTIPLICATION).ir/1:
                        left: BufSamples.ir[0]
                        right: 0.5
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in_]
                -   FFT.kr:
                        buffer_id: LocalBuf.ir[0]
                        source: In.ar[0]
                        hop: 0.75
                        window_type: 1.0
                        active: 1.0
                        window_size: 0.0
                -   PV_MagSmear.kr:
                        pv_chain: FFT.kr[0]
                        bins: 1.0
                -   BufDur.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir/0:
                        left: Control.ir[1:rate]
                        right: BufDur.ir[0]
                -   LFSaw.ar:
                        frequency: BinaryOpUGen(FLOAT_DIVISION).ir/0[0]
                        initial_phase: BinaryOpUGen(SUBTRACTION).ir[0]
                -   MulAdd.ar:
                        source: LFSaw.ar[0]
                        multiplier: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        addend: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                -   BinaryOpUGen(ROUND).ar:
                        left: MulAdd.ar[0]
                        right: 2.0
                -   BufRd.ar:
                        channel_count: 1
                        buffer_id: LocalBuf.ir[0]
                        phase: BinaryOpUGen(ROUND).ar[0]
                        loop: 1.0
                        interpolation: 1.0
                -   BinaryOpUGen(FLOAT_DIVISION).ar:
                        left: BufRd.ar[0]
                        right: Control.ir[0:fft_buffer_size]
                -   UnaryOpUGen(AMPLITUDE_TO_DB).ar:
                        source: BinaryOpUGen(FLOAT_DIVISION).ar[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir/1:
                        left: Control.ir[0:fft_buffer_size]
                        right: Control.ir[1:rate]
                -   ScopeOut2.ar:
                        scope_id: Control.ir[2:scope_id]
                        max_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        scope_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        source[0]: UnaryOpUGen(AMPLITUDE_TO_DB).ar[0]
            """,
        ),
        (
            frequency_scope_log_shm_2,
            """
            synthdef:
                name: supriya:freq-scope-log-shm:2
                ugens:
                -   Control.ir:
                        fft_buffer_size: 2048.0
                        rate: 4.0
                        scope_id: 0.0
                -   UnaryOpUGen(RECIPROCAL).ir:
                        source: Control.ir[0:fft_buffer_size]
                -   BinaryOpUGen(MULTIPLICATION).ir/0:
                        left: Control.ir[1:rate]
                        right: UnaryOpUGen(RECIPROCAL).ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir:
                        left: 1.0
                        right: BinaryOpUGen(MULTIPLICATION).ir/0[0]
                -   Control.kr:
                        in_: 0.0
                -   MaxLocalBufs.ir:
                        maximum: 1.0
                -   LocalBuf.ir:
                        channel_count: 1.0
                        frame_count: Control.ir[0:fft_buffer_size]
                -   BufSamples.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(MULTIPLICATION).ir/1:
                        left: BufSamples.ir[0]
                        right: 0.5
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in_]
                -   FFT.kr:
                        buffer_id: LocalBuf.ir[0]
                        source: In.ar[0]
                        hop: 0.75
                        window_type: 1.0
                        active: 1.0
                        window_size: 0.0
                -   PV_MagSmear.kr:
                        pv_chain: FFT.kr[0]
                        bins: 1.0
                -   BufDur.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir/0:
                        left: Control.ir[1:rate]
                        right: BufDur.ir[0]
                -   LFSaw.ar:
                        frequency: BinaryOpUGen(FLOAT_DIVISION).ir/0[0]
                        initial_phase: BinaryOpUGen(SUBTRACTION).ir[0]
                -   MulAdd.ar:
                        source: LFSaw.ar[0]
                        multiplier: 0.5
                        addend: 0.5
                -   BinaryOpUGen(POWER).ar:
                        left: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        right: MulAdd.ar[0]
                -   BinaryOpUGen(MULTIPLICATION).ar:
                        left: BinaryOpUGen(POWER).ar[0]
                        right: 2.0
                -   BinaryOpUGen(ROUND).ar:
                        left: BinaryOpUGen(MULTIPLICATION).ar[0]
                        right: 2.0
                -   BufRd.ar:
                        channel_count: 1
                        buffer_id: LocalBuf.ir[0]
                        phase: BinaryOpUGen(ROUND).ar[0]
                        loop: 1.0
                        interpolation: 1.0
                -   BinaryOpUGen(FLOAT_DIVISION).ar:
                        left: BufRd.ar[0]
                        right: Control.ir[0:fft_buffer_size]
                -   UnaryOpUGen(AMPLITUDE_TO_DB).ar:
                        source: BinaryOpUGen(FLOAT_DIVISION).ar[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir/1:
                        left: Control.ir[0:fft_buffer_size]
                        right: Control.ir[1:rate]
                -   ScopeOut2.ar:
                        scope_id: Control.ir[2:scope_id]
                        max_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        scope_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        source[0]: UnaryOpUGen(AMPLITUDE_TO_DB).ar[0]
            """,
        ),
    ],
)
def test_supriya(
    synthdef: SynthDef,
    expected_str: str,
) -> None:
    assert normalize(str(synthdef)) == normalize(expected_str)


@pytest.mark.parametrize(
    "sclang_name, sclang_body, sclang_rates, expected_str",
    [
        (
            "system_link_audio_1",
            r"""
            | out=0, in=16, vol=1, level=1, lag=0.05, doneAction=2 |
            var env = EnvGate(i_level: 0, doneAction:doneAction, curve:'sin')
            * Lag.kr(vol * level, lag);
            Out.ar(out, InFeedback.ar(in, 1) * env)
            """,
            r"[\kr, \kr, \kr, \kr, \kr, \ir]",
            """
            synthdef:
                name: system_link_audio_1
                ugens:
                -   Control.ir:
                        doneAction: 2.0
                -   Control.kr:
                        out: 0.0
                        in: 16.0
                        vol: 1.0
                        level: 1.0
                        lag: 0.05000000074505806
                -   BinaryOpUGen(MULTIPLICATION).kr/0:
                        left: Control.kr[2:vol]
                        right: Control.kr[3:level]
                -   Lag.kr:
                        source: BinaryOpUGen(MULTIPLICATION).kr/0[0]
                        lag_time: Control.kr[4:lag]
                -   InFeedback.ar:
                        channel_count: 1
                        bus: Control.kr[1:in]
                -   Control.kr:
                        gate: 1.0
                -   Control.kr:
                        fadeTime: 0.019999999552965164
                -   EnvGen.kr:
                        gate: Control.kr[0:gate]
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: Control.kr[0:fadeTime]
                        done_action: Control.ir[0:doneAction]
                        envelope[0]: 0.0
                        envelope[1]: 2.0
                        envelope[2]: 1.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 1.0
                        envelope[6]: 3.0
                        envelope[7]: 0.0
                        envelope[8]: 0.0
                        envelope[9]: 1.0
                        envelope[10]: 3.0
                        envelope[11]: 0.0
                -   BinaryOpUGen(MULTIPLICATION).kr/1:
                        left: EnvGen.kr[0]
                        right: Lag.kr[0]
                -   BinaryOpUGen(MULTIPLICATION).ar:
                        left: InFeedback.ar[0]
                        right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
                -   Out.ar:
                        bus: Control.kr[0:out]
                        source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
            """,
        ),
        (
            "system_link_audio_2",
            r"""
            | out=0, in=16, vol=1, level=1, lag=0.05, doneAction=2 |
            var env = EnvGate(i_level: 0, doneAction:doneAction, curve:'sin')
            * Lag.kr(vol * level, lag);
            Out.ar(out, InFeedback.ar(in, 2) * env)
            """,
            r"[\kr, \kr, \kr, \kr, \kr, \ir]",
            """
            synthdef:
                name: system_link_audio_2
                ugens:
                -   Control.ir:
                        doneAction: 2.0
                -   Control.kr:
                        out: 0.0
                        in: 16.0
                        vol: 1.0
                        level: 1.0
                        lag: 0.05000000074505806
                -   BinaryOpUGen(MULTIPLICATION).kr/0:
                        left: Control.kr[2:vol]
                        right: Control.kr[3:level]
                -   Lag.kr:
                        source: BinaryOpUGen(MULTIPLICATION).kr/0[0]
                        lag_time: Control.kr[4:lag]
                -   InFeedback.ar:
                        channel_count: 2
                        bus: Control.kr[1:in]
                -   Control.kr:
                        gate: 1.0
                -   Control.kr:
                        fadeTime: 0.019999999552965164
                -   EnvGen.kr:
                        gate: Control.kr[0:gate]
                        level_scale: 1.0
                        level_bias: 0.0
                        time_scale: Control.kr[0:fadeTime]
                        done_action: Control.ir[0:doneAction]
                        envelope[0]: 0.0
                        envelope[1]: 2.0
                        envelope[2]: 1.0
                        envelope[3]: -99.0
                        envelope[4]: 1.0
                        envelope[5]: 1.0
                        envelope[6]: 3.0
                        envelope[7]: 0.0
                        envelope[8]: 0.0
                        envelope[9]: 1.0
                        envelope[10]: 3.0
                        envelope[11]: 0.0
                -   BinaryOpUGen(MULTIPLICATION).kr/1:
                        left: EnvGen.kr[0]
                        right: Lag.kr[0]
                -   BinaryOpUGen(MULTIPLICATION).ar/0:
                        left: InFeedback.ar[0]
                        right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
                -   BinaryOpUGen(MULTIPLICATION).ar/1:
                        left: InFeedback.ar[1]
                        right: BinaryOpUGen(MULTIPLICATION).kr/1[0]
                -   Out.ar:
                        bus: Control.kr[0:out]
                        source[0]: BinaryOpUGen(MULTIPLICATION).ar/0[0]
                        source[1]: BinaryOpUGen(MULTIPLICATION).ar/1[0]
            """,
        ),
        (
            "system_freqScope0_shm",
            r"""
            arg in=0, fftBufSize = 2048, scopebufnum=1, rate=4, dbFactor = 0.02;
			var phase = 1 - (rate * fftBufSize.reciprocal);
			var signal, chain, phasor, numSamples, mul;
			var fftbufnum = LocalBuf(fftBufSize, 1);
			mul = 0.00285;
			numSamples = (BufSamples.ir(fftbufnum) - 2) * 0.5; // 1023 (bufsize=2048)
			signal = In.ar(in);
			chain = FFT(fftbufnum, signal, hop: 0.75, wintype:1);
			chain = PV_MagSmear(chain, 1);
			// -1023 to 1023, 0 to 2046, 2 to 2048 (skip first 2 elements DC and Nyquist)
			phasor = LFSaw.ar(rate/BufDur.ir(fftbufnum), phase, numSamples, numSamples + 2);
			phasor = phasor.round(2); // the evens are magnitude
			ScopeOut2.ar(
				((BufRd.ar(1, fftbufnum, phasor, 1, 1) * mul).ampdb * dbFactor) + 1,
				scopebufnum,
				fftBufSize/rate
			);
            """,
            r"[\kr, \ir, \ir, \ir, \kr]",
            """
            synthdef:
                name: system_freqScope0_shm
                ugens:
                -   Control.ir:
                        fftBufSize: 2048.0
                        scopebufnum: 1.0
                        rate: 4.0
                -   UnaryOpUGen(RECIPROCAL).ir:
                        source: Control.ir[0:fftBufSize]
                -   BinaryOpUGen(MULTIPLICATION).ir/0:
                        left: Control.ir[2:rate]
                        right: UnaryOpUGen(RECIPROCAL).ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir/0:
                        left: 1.0
                        right: BinaryOpUGen(MULTIPLICATION).ir/0[0]
                -   Control.kr:
                        in: 0.0
                        dbFactor: 0.019999999552965164
                -   MaxLocalBufs.ir:
                        maximum: 1.0
                -   LocalBuf.ir:
                        channel_count: 1.0
                        frame_count: Control.ir[0:fftBufSize]
                -   BufSamples.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(SUBTRACTION).ir/1:
                        left: BufSamples.ir[0]
                        right: 2.0
                -   BinaryOpUGen(MULTIPLICATION).ir/1:
                        left: BinaryOpUGen(SUBTRACTION).ir/1[0]
                        right: 0.5
                -   In.ar:
                        channel_count: 1
                        bus: Control.kr[0:in]
                -   FFT.kr:
                        buffer_id: LocalBuf.ir[0]
                        source: In.ar[0]
                        hop: 0.75
                        window_type: 1.0
                        active: 1.0
                        window_size: 0.0
                -   PV_MagSmear.kr:
                        pv_chain: FFT.kr[0]
                        bins: 1.0
                -   BufDur.ir:
                        buffer_id: LocalBuf.ir[0]
                -   BinaryOpUGen(FLOAT_DIVISION).ir/0:
                        left: Control.ir[2:rate]
                        right: BufDur.ir[0]
                -   LFSaw.ar:
                        frequency: BinaryOpUGen(FLOAT_DIVISION).ir/0[0]
                        initial_phase: BinaryOpUGen(SUBTRACTION).ir/0[0]
                -   BinaryOpUGen(ADDITION).ir:
                        left: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        right: 2.0
                -   MulAdd.ar/0:
                        source: LFSaw.ar[0]
                        multiplier: BinaryOpUGen(MULTIPLICATION).ir/1[0]
                        addend: BinaryOpUGen(ADDITION).ir[0]
                -   BinaryOpUGen(ROUND).ar:
                        left: MulAdd.ar/0[0]
                        right: 2.0
                -   BufRd.ar:
                        channel_count: 1
                        buffer_id: LocalBuf.ir[0]
                        phase: BinaryOpUGen(ROUND).ar[0]
                        loop: 1.0
                        interpolation: 1.0
                -   BinaryOpUGen(MULTIPLICATION).ar:
                        left: BufRd.ar[0]
                        right: 0.002850000048056245
                -   UnaryOpUGen(AMPLITUDE_TO_DB).ar:
                        source: BinaryOpUGen(MULTIPLICATION).ar[0]
                -   MulAdd.ar/1:
                        source: UnaryOpUGen(AMPLITUDE_TO_DB).ar[0]
                        multiplier: Control.kr[1:dbFactor]
                        addend: 1.0
                -   BinaryOpUGen(FLOAT_DIVISION).ir/1:
                        left: Control.ir[0:fftBufSize]
                        right: Control.ir[2:rate]
                -   ScopeOut2.ar:
                        scope_id: Control.ir[1:scopebufnum]
                        max_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        scope_frames: BinaryOpUGen(FLOAT_DIVISION).ir/1[0]
                        source[0]: MulAdd.ar/1[0]
            """,
        ),
    ],
)
@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_sclang(
    sclang_name: str, sclang_body: str, sclang_rates: str, expected_str: str
) -> None:
    compiled = SuperColliderSynthDef(sclang_name, sclang_body, sclang_rates).compile()
    synthdef = decompile_synthdef(compiled)
    assert normalize(str(synthdef)) == normalize(expected_str)



from threading import Lock
from typing import TYPE_CHECKING, Literal, cast

from ..enums import AddAction, CalculationRate, ServerLifecycleEvent
from ..exceptions import ServerOffline
from ..typing import AddActionLike
from ..ugens import SynthDef
from ..ugens.system import AMPLITUDE_SCOPE_SYNTHDEFS, FREQUENCY_SCOPE_SYNTHDEFS
from .entities import Bus, BusGroup, Node, ScopeBuffer, Synth

if TYPE_CHECKING:
    from .realtime import BaseServer, ServerLifecycleCallback


class BaseScope:
    """
    Base class for scopes.
    """

    def __init__(
        self,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_TAIL,
        bus: Bus | BusGroup,
        context: "BaseServer",
        target_node: Node | None = None,
    ) -> None:
        if bus.context is not context:
            raise ValueError(bus, context)
        elif target_node and target_node.context is not context:
            raise ValueError(target_node, context)
        self.add_action = AddAction.from_expr(add_action)
        self.bus = bus
        self.channel_count = 0
        self.context = context
        self.lifecycle_callback: ServerLifecycleCallback | None = None
        self.lock = Lock()
        self.max_frames = 0
        self.scope_buffer: ScopeBuffer | None = None
        self.status: Literal["online", "offline"] = "offline"
        self.synth: Synth | None = None
        self.target_node: Node = target_node or self.context.root_node

    def _add_synth(self, scope_buffer: ScopeBuffer, synthdef: SynthDef) -> Synth:
        raise NotImplementedError

    def _get_synthdef(self) -> SynthDef:
        raise NotImplementedError

    def play(self) -> None:
        """
        Start the scope.
        """
        from .realtime import Server

        with self.lock:
            if self.status == "online":
                return
            if self.context._shared_memory is None:
                raise ValueError
            self.scope_buffer = self.context.add_scope_buffer()
            synthdef = self._get_synthdef()
            with self.context.at():
                with self.context.add_synthdefs(synthdef):
                    self.synth = self._add_synth(self.scope_buffer, synthdef)
            self.lifecycle_callback = cast(
                Server, self.context
            ).register_lifecycle_callback(
                event=[
                    ServerLifecycleEvent.OSC_PANICKED,
                    ServerLifecycleEvent.PROCESS_PANICKED,
                    ServerLifecycleEvent.QUITTING,
                ],
                once=True,
                procedure=lambda event: self.stop(),
            )
            self.status = "online"

    def read(self) -> tuple[int, list[float]]:
        """
        Read latest data from an online scope.
        """
        if not self.status == "online":
            raise ValueError
        elif self.context._shared_memory is None:
            raise ValueError
        elif self.scope_buffer is None:
            raise ValueError
        self.channel_count, self.max_frames = (
            self.context._shared_memory.describe_scope_buffer(int(self.scope_buffer))
        )
        return self.context._shared_memory.read_scope_buffer(int(self.scope_buffer))

    def stop(self) -> None:
        """
        Stop the scope.
        """
        with self.lock:
            if self.status == "offline":
                return
            if self.synth:
                try:
                    self.synth.free()
                except ServerOffline:
                    pass
                self.synth = None
            if self.scope_buffer:
                try:
                    self.scope_buffer.free()
                except ServerOffline:
                    pass
                self.scope_buffer = None
            if self.lifecycle_callback:
                self.context.unregister_lifecycle_callback(self.lifecycle_callback)
                self.lifecycle_callback = None
            self.status = "offline"


class AmplitudeScope(BaseScope):
    """
    An amplitude scope.
    """

    def _add_synth(self, scope_buffer: ScopeBuffer, synthdef: SynthDef) -> Synth:
        return self.context.add_synth(
            add_action=self.add_action,
            in_=int(self.bus),
            scope_id=int(scope_buffer),
            synthdef=synthdef,
            target_node=self.target_node,
        )

    def _get_synthdef(self) -> SynthDef:
        token = self.bus.calculation_rate.token
        if (
            channel_count := len(self.bus) if isinstance(self.bus, BusGroup) else 1
        ) > 16:
            raise ValueError(self.bus)
        return AMPLITUDE_SCOPE_SYNTHDEFS[f"supriya:amp-scope-{token}:{channel_count}"]


class FrequencyScope(BaseScope):
    """
    A frequency scope.
    """

    def __init__(
        self,
        *,
        add_action: AddActionLike = AddAction.ADD_TO_TAIL,
        bus: Bus | BusGroup,
        context: "BaseServer",
        fft_size: int = 4096,
        frequency_mode: Literal["linear", "logarithmic"] = "linear",
        rate: int = 4,
        target_node: Node | None = None,
    ) -> None:
        super().__init__(
            add_action=add_action,
            bus=bus,
            context=context,
            target_node=target_node,
        )
        if frequency_mode not in ("linear", "logarithmic"):
            raise ValueError(frequency_mode)
        elif self.bus.calculation_rate != CalculationRate.AUDIO:
            raise ValueError(bus.calculation_rate)
        elif fft_size < 256 or fft_size > 8192:
            raise ValueError(fft_size)
        self.fft_size = fft_size
        self.frequency_mode = frequency_mode
        self.rate = rate

    def _add_synth(self, scope_buffer: ScopeBuffer, synthdef: SynthDef) -> Synth:
        return self.context.add_synth(
            add_action=self.add_action,
            fft_buffer_size=self.fft_size,
            in_=int(self.bus),
            rate=self.rate,
            scope_id=int(scope_buffer),
            synthdef=synthdef,
            target_node=self.target_node,
        )

    def _get_synthdef(self) -> SynthDef:
        return FREQUENCY_SCOPE_SYNTHDEFS[
            f"supriya:freq-scope-{self.frequency_mode[:3]}-shm:1"
        ]



import pytest

from supriya import AddAction, CalculationRate
from supriya.patterns import (
    BusAllocateEvent,
    BusFreeEvent,
    BusPattern,
    CompositeEvent,
    Event,
    EventPattern,
    GroupAllocateEvent,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    ParallelPattern,
    SequencePattern,
    SynthAllocateEvent,
)
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test
from supriya.typing import CalculationRateLike
from supriya.ugens import system


@pytest.mark.parametrize(
    "stop_at, inner_pattern, calculation_rate, channel_count, release_time, expected, is_infinite",
    [
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.0,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.0,
                            in_=M("A"),
                            synthdef=system.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                NoteEvent(M("D"), a=1, out=M("A"), target_node=M("B")),
                NoteEvent(M("E"), a=2, out=M("A"), target_node=M("B")),
                CompositeEvent(
                    [NodeFreeEvent(M("C")), NodeFreeEvent(M("B")), BusFreeEvent(M("A"))]
                ),
            ],
            False,
        ),
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("A"),
                            synthdef=system.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                NoteEvent(M("D"), a=1, out=M("A"), target_node=M("B")),
                NoteEvent(M("E"), a=2, out=M("A"), target_node=M("B")),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("C")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("B")),
                        BusFreeEvent(M("A")),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            BusPattern(EventPattern(a=SequencePattern([1, 2])), channel_count=2),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("A"),
                            synthdef=system.system_link_audio_2,
                            target_node=M("B"),
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("D"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("E"), target_node=M("B")),
                        SynthAllocateEvent(
                            M("F"),
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("D"),
                            out=M("A"),
                            synthdef=system.system_link_audio_2,
                            target_node=M("E"),
                        ),
                    ]
                ),
                NoteEvent(M("G"), a=1, out=M("D"), target_node=M("E")),
                NoteEvent(M("H"), a=2, out=M("D"), target_node=M("E")),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("F")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("E")),
                        BusFreeEvent(M("D")),
                    ]
                ),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("C")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("B")),
                        BusFreeEvent(M("A")),
                    ]
                ),
            ],
            False,
        ),
        (
            None,
            ParallelPattern(
                [
                    BusPattern(
                        EventPattern(a=SequencePattern([1, 2])), channel_count=2
                    ),
                    BusPattern(
                        EventPattern(a=SequencePattern([1, 2])), channel_count=2
                    ),
                ]
            ),
            "audio",
            2,
            0.25,
            [
                CompositeEvent(
                    [
                        BusAllocateEvent(
                            M("A"),
                            calculation_rate=CalculationRate.AUDIO,
                            channel_count=2,
                        ),
                        GroupAllocateEvent(M("B")),
                        SynthAllocateEvent(
                            M("C"),
                            system.system_link_audio_2,
                            add_action=AddAction.ADD_AFTER,
                            amplitude=1.0,
                            fade_time=0.25,
                            in_=M("A"),
                            target_node=M("B"),
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        CompositeEvent(
                            [
                                BusAllocateEvent(
                                    M("D"),
                                    calculation_rate=CalculationRate.AUDIO,
                                    channel_count=2,
                                ),
                                GroupAllocateEvent(M("E"), target_node=M("B")),
                                SynthAllocateEvent(
                                    M("F"),
                                    system.system_link_audio_2,
                                    add_action=AddAction.ADD_AFTER,
                                    amplitude=1.0,
                                    fade_time=0.25,
                                    in_=M("D"),
                                    out=M("A"),
                                    target_node=M("E"),
                                ),
                            ]
                        ),
                        NoteEvent(
                            M("G"),
                            a=1,
                            delta=0.0,
                            out=M("D"),
                            target_node=M("E"),
                        ),
                        CompositeEvent(
                            [
                                BusAllocateEvent(
                                    M("H"),
                                    calculation_rate=CalculationRate.AUDIO,
                                    channel_count=2,
                                ),
                                GroupAllocateEvent(M("I"), target_node=M("B")),
                                SynthAllocateEvent(
                                    M("J"),
                                    system.system_link_audio_2,
                                    add_action=AddAction.ADD_AFTER,
                                    amplitude=1.0,
                                    fade_time=0.25,
                                    in_=M("H"),
                                    out=M("A"),
                                    target_node=M("I"),
                                ),
                            ]
                        ),
                        NoteEvent(
                            M("K"),
                            a=1,
                            delta=0.0,
                            out=M("H"),
                            target_node=M("I"),
                        ),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        NoteEvent(
                            M("L"),
                            a=2,
                            delta=0.0,
                            out=M("D"),
                            target_node=M("E"),
                        ),
                        NoteEvent(
                            M("M"),
                            a=2,
                            delta=0.0,
                            out=M("H"),
                            target_node=M("I"),
                        ),
                    ],
                    delta=1.0,
                ),
                CompositeEvent(
                    [
                        CompositeEvent(
                            [
                                NodeFreeEvent(M("F")),
                                NullEvent(delta=0.25),
                                NodeFreeEvent(M("E")),
                                BusFreeEvent(M("D")),
                            ]
                        ),
                        CompositeEvent(
                            [
                                NodeFreeEvent(M("J")),
                                NullEvent(delta=0.25),
                                NodeFreeEvent(M("I")),
                                BusFreeEvent(M("H")),
                            ]
                        ),
                    ]
                ),
                CompositeEvent(
                    [
                        NodeFreeEvent(M("C")),
                        NullEvent(delta=0.25),
                        NodeFreeEvent(M("B")),
                        BusFreeEvent(M("A")),
                    ]
                ),
            ],
            False,
        ),
    ],
)
def test_pattern(
    stop_at: float | None,
    inner_pattern,
    calculation_rate: CalculationRateLike,
    channel_count: int,
    release_time: float,
    expected: list[Event],
    is_infinite: bool,
) -> None:
    pattern = BusPattern(
        inner_pattern,
        calculation_rate=calculation_rate,
        channel_count=channel_count,
        release_time=release_time,
    )
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import pytest

from supriya.patterns import ChoicePattern, Pattern, SeedPattern, SequencePattern
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, pattern, expected, is_infinite",
    [
        (None, SequencePattern([1, 2, 3, 4]), [1, 2, 3, 4], False),
        (1, SequencePattern([1, 2, 3, 4]), [1], False),
    ],
)
def test_pattern(
    stop_at: float | None, pattern: Pattern, expected: list[float], is_infinite: bool
) -> None:
    run_pattern_test(SeedPattern(pattern), expected, is_infinite, stop_at)


def test_random() -> None:
    choice_pattern = ChoicePattern([1, 2, 3])
    assert len(set(tuple(choice_pattern) for _ in range(10))) > 1
    seed_pattern = SeedPattern(choice_pattern)
    assert len(set(tuple(seed_pattern) for _ in range(10))) == 1



import logging

import pytest

from supriya.clocks import (
    CallbackEvent,
    ClockContext,
    ClockDelta,
    OfflineClock,
    TimeUnit,
)

repeat_count = 5

logger = logging.getLogger("supriya.test")


@pytest.fixture(autouse=True)
def log_everything(caplog) -> None:
    caplog.set_level(logging.INFO, logger="supriya")


def callback(
    context: ClockContext,
    store: list[ClockContext],
    *,
    blow_up_at: int | None = None,
    delta: float = 0.25,
    limit: int | None = 4,
    time_unit: TimeUnit = TimeUnit.BEATS,
) -> ClockDelta:
    assert isinstance(context.event, CallbackEvent)
    if context.event.invocations == blow_up_at:
        raise Exception
    store.append(context)
    if limit is None:
        return delta, time_unit
    elif context.event.invocations < limit:
        return delta, time_unit
    return None


def check(
    store: list[ClockContext],
) -> list[tuple[list[float | str], list[float | int], list[float | int]]]:
    return [
        (
            [
                "{}/{}".format(*context.current_moment.time_signature),
                context.current_moment.beats_per_minute,
            ],
            [
                context.current_moment.measure,
                round(context.current_moment.measure_offset, 10),
                context.current_moment.offset,
                context.current_moment.seconds,
            ],
            [
                context.desired_moment.measure,
                round(context.desired_moment.measure_offset, 10),
                context.desired_moment.offset,
                context.desired_moment.seconds,
            ],
        )
        for context in store
    ]


def test_basic() -> None:
    clock = OfflineClock()
    store: list[ClockContext] = []
    clock.schedule(callback, schedule_at=0.0, args=[store])
    clock.start()
    assert check(store) == [
        (["4/4", 120.0], [1, 0.0, 0.0, 0.0], [1, 0.0, 0.0, 0.0]),
        (["4/4", 120.0], [1, 0.25, 0.25, 0.5], [1, 0.25, 0.25, 0.5]),
        (["4/4", 120.0], [1, 0.5, 0.5, 1.0], [1, 0.5, 0.5, 1.0]),
        (["4/4", 120.0], [1, 0.75, 0.75, 1.5], [1, 0.75, 0.75, 1.5]),
        (["4/4", 120.0], [2, 0.0, 1.0, 2.0], [2, 0.0, 1.0, 2.0]),
    ]



import logging

import pytest

from supriya.enums import ParameterRate
from supriya.ugens import Out, Parameter, SinOsc, SynthDef, SynthDefBuilder


@pytest.fixture(autouse=True)
def capture_logs(caplog):
    caplog.set_level(logging.INFO, logger="supriya")


@pytest.fixture
def two_voice_synthdef() -> SynthDef:
    with SynthDefBuilder(
        frequencies=(220, 440),
        amplitude=Parameter(value=1.0, rate=ParameterRate.AUDIO),
    ) as builder:
        sin_osc = SinOsc.ar(frequency=builder["frequencies"])
        enveloped_sin = sin_osc * builder["amplitude"]
        Out.ar(bus=0, source=enveloped_sin)
    return builder.build(name="test:two-voice")
